unit uTCPSockets;

{$mode objfpc}{$H+}

{$IFDEF WIN32}
  {$MESSAGE WARNING 'There is a limited amount of threads that a 32 bit process(2 GB address space) can create. For better performance please build synapse client/server with a 64 bit version.'}
{$ENDIF}

interface

uses
  Classes, SysUtils, FileUtil, Contnrs, Md5, SyncObjs, Synsock, BlckSock;

const
  Sep = '$%&^@';
  DefPort = 40516;
  DefProxyPort = 8080;
  Timeout = 6000;
  TunnelTimeout = 60000;
  ConnectionTimeout = 5000;
  PingTimeoutClient = 30000;
  PingTimeoutServer = 300000;
  BaseThreadStackSize = 64*1024;
  WorkerThreadStackSize = 4*1024*1024;

type
  TConData = class
    FUniqueID: Integer;
    FDateTime: TDateTime;
    FLocalIP: String;
    FLocalPort: Integer;
    FLocalHost: String;
    FLocalUser: String;
    FRemoteIP: String;
    FRemotePort: Integer;
    FRemoteHost: String;
    FRemoteUser: String;
    FData: Pointer;
  public
    constructor Create;
  end;

  TLock = class
  private
    FCS: TRTLCriticalSection;
    FLocked: Integer;
    function Lock: Boolean;
    procedure Unlock;
  public
    constructor Create; virtual;
    destructor Destroy; override;
  end;

  TLockList = class(TLock)
  private
    FList: TList;
  public
    constructor Create; override;
    destructor Destroy; override;
  end;

  TTCPBase = class;

  TUpdateType = (utAdd, utRemove);
  TThreadType = (ttConnectedClient, ttAcceptedClient, ttConnectedWorker, ttAcceptedWorker);
  TBase = class(TThread)
  private
    FEvent: TEvent;
    FMsg: String;
    FParams: TStringArray;
    FErrMsg: String;
    FErrCode: Integer;
    FUniqueID: Integer;
    FMaxThreadException: Boolean;
    FBlockSocket: TTCPBlockSocket;
    FConData: TConData;
    FStarted: Boolean;
    FThreadType: TThreadType;
    FKey: String;
    FBusy: Boolean;
    FTCPBase: TTCPBase;
    procedure DoOnErrorMessage;
    procedure DoOnSentMessage;
    function GetThreadCnt: Integer;
    function GetExtendedInfo: Boolean;
    procedure ErrorMessage(AErrMsg: String; AErrCode: Integer);
    procedure Wait(AMs: Integer);
    procedure UpdateGUI(AMethod: TThreadMethod; AQueue: Boolean = False);
    procedure UpdateThreadList(AUpdateType: TUpdateType);
    function SendMessage(AMsg: String; AParams: TStringArray; AIsInternal: Boolean = False): Boolean;
    function RecvMessage(ATimeOut: Integer; out AMsg: String; out AParams: TStringArray): Boolean;
  protected
    procedure TerminatedSet; override;
  public
    constructor Create(ATCPBase: TTCPBase; AStackSize: SizeUInt);
    destructor Destroy; override;
  end;

  TConnection = class;

  TWorkDirection = (wdDownload, wdUpload, wdError);
  TWorker = class(TBase)
  private
    FOldCnt: Int64;
    FTick: QWord;
    FMS: TMemoryStream;
    FFileName: String;
    FCnt: Int64;
    FTotCnt: Int64;
    FSpeed: Int64;
    FForceQuit: Boolean;
    FRemaining: Int64;
    FWorkDirection: TWorkDirection;
    FConnection: TConnection;
    function SendFile(AFileName: String): Boolean;
    function RecvFile(AFileName: String; ASize: Int64): Boolean;
    function SendStream(AMS: TMemoryStream): Boolean;
    function RecvStream(AMS: TMemoryStream): Boolean;
    function AddDirSeparator(const Path: String): String;
    function GetDownloadDir(const APath: String): String;
    function ChangeFileName(AFileName, AFrom, ATo: String): String;
    procedure Upload;
    procedure Download;
    procedure DoOnSentFile;
    procedure DoOnRecvFile;
    procedure DoOnFileProgress;
    procedure Progress(AFileName: String; ACnt, ATotCnt: Int64; AError: Boolean = False);
  protected
    procedure Execute; override;
  public
    constructor Create(ATCPBase: TTCPBase; AStackSize: SizeUInt);
    constructor Create(ATCPBase: TTCPBase; AStackSize: SizeUInt; ASocket: TSocket);
    destructor Destroy; override;
  public
    property ForceQuit: Boolean read FForceQuit write FForceQuit;
    property ConData: TConData read FConData write FConData;
  end;

  TOnConnected = procedure(Sender: TObject; AConData: TConData; AConCnt: Integer = 0) of object;
  TOnDisconnected = procedure(Sender: TObject; AConData: TConData; AConCnt: Integer = 0; AByUser: Boolean = False) of object;
  TOnSocketError = procedure(Sender: TObject; AErrMsg: string; AErrCode: Integer; AConData: TConData) of object;
  TOnMessage = procedure(Sender: TObject; AMsg: String; AParams: TStringArray; AConData: TConData) of object;
  TOnStream = procedure(Sender: TObject; AMS: TMemoryStream; AConData: TConData) of object;
  TOnFile =  procedure(Sender: TObject; AFileName: String; AConData: TConData) of object;
  TOnStreamProgress = procedure(Sender: TObject; AWorkDirection: TWorkDirection; ACnt, ATotCnt, ASpeed, ARemaining: Int64; AConData: TConData) of object;
  TOnFileProgress = procedure(Sender: TObject; AWorkDirection: TWorkDirection; AFileName: String; ACnt, ATotCnt, ASpeed, ARemaining: Int64; AConData: TConData) of object;
  TTCPBase = class
  private
    FUniqueID: Integer;
    FIsDaemon: Boolean;
    FDownloadDirectory: String;
    FLastErrMsg: String;
    FLastErrCode: Integer;
    FCompress: Boolean;
    FActiveOrConnected: Boolean;
    FShowInternalMessages: Boolean;
    FShowErrorMessages: Boolean;
    FOnConnected: TOnConnected;
    FOnDisconnected: TOnDisconnected;
    FOnSocketError: TOnSocketError;
    FOnSentMessage: TOnMessage;
    FOnSentStream: TOnStream;
    FOnSentFile: TOnFile;
    FOnRecvMessage: TOnMessage;
    FOnRecvStream: TOnStream;
    FOnRecvFile: TOnFile;
    FOnStreamProgress: TOnStreamProgress;
    FOnFileProgress: TOnFileProgress;
    FThreadList: TLockList;
    procedure CopyParams(ASrc: TStringArray; out ADst: TStringArray);
    procedure CopyStream(ASrc, ADst: TMemoryStream);
  protected
    procedure ErrorMessage(AThread: TBase; AErrMsg: String; AErrCode: Integer; AConData: TConData); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
  public
    property LastErrMsg: String read FLastErrMsg;
    property LastErrCode: Integer read FLastErrCode;
    property IsDaemon: Boolean read FIsDaemon write FIsDaemon;
    property DownloadDirectory: String read FDownloadDirectory write FDownloadDirectory;
    property Compress: Boolean read FCompress write FCompress;
    property OnConnected: TOnConnected read FOnConnected write FOnConnected;
    property OnDisconnected: TOnDisconnected read FOnDisconnected write FOnDisconnected;
    property ShowErrorMessages: Boolean read FShowErrorMessages write FShowErrorMessages;
    property OnSocketError: TOnSocketError read FOnSocketError write FOnSocketError;
    property OnSentMessage: TOnMessage read FOnSentMessage write FOnSentMessage;
    property OnSentStream: TOnStream read FOnSentStream write FOnSentStream;
    property OnSentFile: TOnFile read FOnSentFile write FOnSentFile;
    property OnRecvMessage: TOnMessage read FOnRecvMessage write FOnRecvMessage;
    property OnRecvStream: TOnStream read FOnRecvStream write FOnRecvStream;
    property OnRecvFile: TOnFile read FOnRecvFile write FOnRecvFile;
    property OnStreamProgress: TOnStreamProgress read FOnStreamProgress write FOnStreamProgress;
    property OnFileProgress: TOnFileProgress read FOnFileProgress write FOnFileProgress;
  end;

  TTCPClient = class;

  TProxyType = (ptNone, ptHTML, ptSocks4, ptSocks5);
  TProxy = record
    FProxyType: TProxyType;
    FProxyIP: String;
    FProxyPort: Integer;
    FProxyUser: String;
    FProxyPass: String;
  end;

  TConnection = class
    FIP: String;
    FPort: Integer;
    FUser: String;
    FPass: String;
    FProxy: TProxy;
    FData: Pointer;
  public
    constructor Create;
  end;

  TConnectedClient = class(TBase)
  private
    FLastPing: QWord;
    FForceQuit: Boolean;
    FTCPClient: TTCPClient;
    procedure DoOnConnected;
    procedure DoOnDisconnected;
    procedure DoOnRecvMessage;
    procedure DoOnFileRequest;
  protected
    procedure Execute; override;
  public
    constructor Create(ATCPBase: TTCPBase; AStackSize: SizeUInt);
    destructor Destroy; override;
  public
    property ForceQuit: Boolean read FForceQuit write FForceQuit;
    property ConData: TConData read FConData;
  end;

  TTCPClient = class(TTCPBase)
  private
    FBusy: Boolean;
    FUseProxy: Boolean;
    FCC: TConnectedClient;
    FConnection: TConnection;
    function GetConnected: Boolean;
    function Connect(ABase: TBase; AConnection: TConnection): Boolean;
    function Login(ABase: TBase; AConnection: TConnection): Boolean;
    function StartWorker(AWorkDirection: TWorkDirection; AFileName: String; AData: Pointer): Boolean;
    procedure SetupProxy;
    procedure CopyConnection(ASrc, ADst: TConnection);
    procedure DoOnTerminate(Sender: TObject);
    procedure FileRequest(AFileName: String);
  protected
    procedure ErrorMessage(AThread: TBase; AErrMsg: String; AErrCode: Integer; AConData: TConData); override;
  public
    constructor Create; override;
    destructor Destroy; override;
  public
    procedure Connect(const AConnection: TConnection);
    procedure Disconnect;
    function SendMessage(AMsg: String; AParams: TStringArray; AData: Pointer = nil): Boolean;
    function SendFile(AFileName: String; AData: Pointer = nil): Boolean;
  public
    property Connected: Boolean read GetConnected;
    property CC: TConnectedClient read FCC;
  end;

  TTCPServer = class;

  PUserData = ^TUserData;
  TUserData = record
    FName: String;
    FIP: String;
  end;

  TBind = class
    FIP: String;
    FPort: Integer;
    FUser: String;
    FPass: String;
    FData: Pointer;
  public
    constructor Create;
  end;

  TAcceptedClient = class(TBase)
  private
    FLastPing: QWord;
    FTCPServer: TTCPServer;
    FForceQuit: Boolean;
    FConCnt: Integer;
    function Login: Byte;
    function AddUser(AUserName: String): Boolean;
    function RemoveUser(AUserName: String): Boolean;
    procedure StartWorker;
    procedure DoOnConnected;
    procedure DoOnDisconnected;
    procedure DoOnRecvMessage;
  protected
    procedure Execute; override;
  public
    constructor Create(ATCPBase: TTCPBase; AStackSize: SizeUInt; ASocket: TSocket);
    destructor Destroy; override;
  public
    property ForceQuit: Boolean read FForceQuit write FForceQuit;
    property ConData: TConData read FConData;
  end;

  TThreadManager = class(TBase)
  private
    FTCPServer: TTCPServer;
  protected
    procedure Execute; override;
  public
    constructor Create(ATCPBase: TTCPBase);
  end;

  TTCPServer = class(TTCPBase)
  private
    FBusy: Boolean;
    FTM: TThreadManager;
    FBind: TBind;
    FLock: TLock;
    FUserList: TFPHashList;
    function GetActive: Boolean;
    procedure ClearUserList;
    procedure CopyBind(ASrc, ADst: TBind);
    procedure CopyConnectionData(ASrc, ADst: TConData);
  protected
    procedure ErrorMessage(AThread: TBase; AErrMsg: String; AErrCode: Integer; AConData: TConData); override;
  public
    constructor Create; override;
    destructor Destroy; override;
  public
    procedure Start(ABindCon: TBind);
    procedure Stop;
    function SendMessage(AAC: TAcceptedClient; AMsg: String; AParams: TStringArray): Boolean;
    function SendFile(AAC: TAcceptedClient; AFileName: String): Boolean;
  public
    property Active: Boolean read GetActive;
  end;

implementation

uses uConst, uCryptoCompression;

{ TConData }
constructor TConData.Create;
begin
  FUniqueID := 0;
  FDateTime := Now;
  FLocalIP := '';
  FLocalPort := DefPort;
  FLocalHost := '';
  FLocalUser := '';
  FRemoteIP := '';
  FRemotePort := DefPort;
  FRemoteHost := '';
  FRemoteUser := '';
  FData := nil;
end;

{ TLock }
constructor TLock.Create;
begin
  FLocked := 0;
  InitCriticalSection(FCS);
end;

destructor TLock.Destroy;
begin
  DoneCriticalSection(FCS);
  inherited Destroy;
end;

function TLock.Lock: Boolean;
begin
  Result := False;
  if FLocked = 1 then
    Exit;
  FLocked := 1;
  EnterCriticalSection(FCS);
  Result := True;
end;

procedure TLock.Unlock;
begin
  LeaveCriticalSection(FCS);
  FLocked := 0;
end;

{ LockList }
constructor TLockList.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TLockList.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

{ TBase }
constructor TBase.Create(ATCPBase: TTCPBase; AStackSize: SizeUInt);
begin
  FMaxThreadException := False;
  try
    inherited Create(True, AStackSize);
  except
    on E: EThread do
    begin
      FMaxThreadException := True;
      ATCPBase.ErrorMessage(nil, rsThreadError, -1, nil);
      Exit;
    end;
  end;
  FMsg := '';
  FErrMsg := '';
  FErrCode := 0;
  FUniqueID := 0;
  FStarted := False;
  FTCPBase := ATCPBase;
  FKey := SHA1Hash(WorkerThreadStackSize.ToString);
  FEvent := TEvent.Create(nil, True, False, '');
  FConData := TConData.Create;
  FBlockSocket := TTCPBlockSocket.Create;
  FBlockSocket.ConnectionTimeout := ConnectionTimeout;
  FBlockSocket.Family := SF_Any;
  FBlockSocket.RaiseExcept := False;
end;

destructor TBase.Destroy;
begin
  if FMaxThreadException then
    Exit;
  FEvent.Free;
  FBlockSocket.Free;
  FConData.Free;
  inherited Destroy;
end;

procedure TBase.DoOnErrorMessage;
begin
  FTCPBase.ErrorMessage(Self, FErrMsg, FErrCode, FConData);
end;

procedure TBase.DoOnSentMessage;
begin
  if Assigned(FTCPBase.FOnSentMessage) then
    FTCPBase.FOnSentMessage(Self, FMsg, FParams, FConData);
end;

function TBase.GetThreadCnt: Integer;
begin
  Result := -1;
  if FTCPBase.FThreadList.Lock then
  try
    Result := FTCPBase.FThreadList.FList.Count;
  finally
    FTCPBase.FThreadList.Unlock;
  end;
end;

function TBase.GetExtendedInfo: Boolean;
begin
  Result := True;
  FBlockSocket.GetSins;
  if FBlockSocket.LastError <> 0 then
  begin
    Result := False;
    FTCPBase.ErrorMessage(Self, FBlockSocket.LastErrorDesc, FBlockSocket.LastError, FConData);
    Exit;
  end;
  with FConData do
  begin
    FLocalIP := FBlockSocket.GetLocalSinIP;
    FLocalPort := FBlockSocket.GetLocalSinPort;
    FLocalHost := FBlockSocket.ResolveIPToName(FLocalIP);
    FRemoteIP := FBlockSocket.GetRemoteSinIP;
    FRemotePort := FBlockSocket.GetRemoteSinPort;
    FRemoteHost := FBlockSocket.ResolveIPToName(FRemoteIP);
  end;
end;

procedure TBase.TerminatedSet;
begin
  if FStarted then
    FEvent.SetEvent;
end;

procedure TBase.ErrorMessage(AErrMsg: String; AErrCode: Integer);
begin
  if (not FTCPBase.FActiveOrConnected) or (not FTCPBase.FShowErrorMessages) then
    Exit;
  FErrMsg := AErrMsg;
  FErrCode := AErrCode;
  UpdateGUI(@DoOnErrorMessage);
end;

procedure TBase.UpdateGUI(AMethod: TThreadMethod; AQueue: Boolean);
begin
  if FTCPBase.FIsDaemon then
    AMethod
  else
  begin
    if AQueue then
      Queue(AMethod)
    else
      Synchronize(AMethod);
  end;
end;

procedure TBase.UpdateThreadList(AUpdateType: TUpdateType);
begin
  while (not Terminated) do
  begin
    if FTCPBase.FThreadList.Lock then
    begin
      try
        case AUpdateType of
          utAdd:
            begin
              Inc(FTCPBase.FUniqueID);
              FConData.FUniqueID := FTCPBase.FUniqueID;
              FConData.FDateTime := Now;
              FUniqueID := FTCPBase.FUniqueID;
              FTCPBase.FThreadList.FList.Add(Self);
            end;
          utRemove:
            FTCPBase.FThreadList.FList.Remove(Self);
        end;
      finally
        FTCPBase.FThreadList.Unlock;
      end;
      Break;
    end;
    Wait(Random(100));
  end;
end;

procedure TBase.Wait(AMs: Integer);
begin
  FEvent.WaitFor(AMs);
end;

function TBase.SendMessage(AMsg: String; AParams: TStringArray; AIsInternal: Boolean): Boolean;
var
  I: Integer;
begin
  Result := False;
  if (not FTCPBase.FActiveOrConnected) or (FBlockSocket.Socket = INVALID_SOCKET) then
    Exit;
  FBusy := True;
  try
    FMsg := AMsg;
    FTCPBase.CopyParams(AParams, FParams);
    if Length(AParams) > 0 then
    begin
      AMsg := Trim(AMsg) + Sep;
      for I := Low(AParams) to High(AParams) do
        AMsg := AMsg + Trim(AParams[I]) + Sep;
    end;
    AMsg := Encrypt(FKey, AMsg);
    FBlockSocket.SendString(AMsg + #13#10);
    Result := FBlockSocket.LastError = 0;
    if Result then
    begin
      if not AIsInternal then
        UpdateGUI(@DoOnSentMessage)
    end
    else
      ErrorMessage(Format(rsMessageSentError, [FMsg]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
  finally
    FBusy := False;
  end;
end;

function TBase.RecvMessage(ATimeOut: Integer; out AMsg: String;
  out AParams: TStringArray): Boolean;
var
  Str: String;
  P: Integer;
  Len: Integer;
begin
  Result := True;
  if (not FTCPBase.FActiveOrConnected) or (FBlockSocket.Socket = INVALID_SOCKET) then
    Exit;

  FBusy := True;
  try
    AMsg := '';
    AParams := nil;
    Str := FBlockSocket.RecvString(ATimeOut);
    if (FBlockSocket.LastError > 0) and (FBlockSocket.LastError <> WSAETIMEDOUT) then
    begin
      Result := False;
      ErrorMessage(FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
      Exit;
    end;
    if Trim(Str) <> '' then
    begin
      Str := Decrypt(FKey, Str);
      P := Pos(Sep, Str);
      if P > 0 then
      begin
        AMsg := Trim(System.Copy(Str, 1, P - 1));
        System.Delete(Str, 1, P - 1 + Length(Sep));
        repeat
          P := Pos(Sep, Str);
          if P > 0 then
          begin
            Len := Length(AParams);
            SetLength(AParams, Len + 1);
            AParams[Len] := Trim(System.Copy(Str, 1, P - 1));
            System.Delete(Str, 1, P - 1 + Length(Sep));
          end;
        until (P = 0) or (Str = '') ;
      end
      else
        AMsg := Trim(Str);
    end;
  finally
    FBusy := False;
  end;
end;

{ TWorker}
constructor TWorker.Create(ATCPBase: TTCPBase; AStackSize: SizeUInt);
begin
  inherited Create(ATCPBase, AStackSize);
  if FMaxThreadException then
    Exit;
  Randomize;
  FForceQuit := False;
  FCnt := 0;
  FTotCnt := 0;
  FSpeed := 0;
  FRemaining := 0;
  FFileName := '';
  FMS := TMemoryStream.Create;
  FTCPBase := ATCPBase;
  FConnection := TConnection.Create;
  FThreadType := ttConnectedWorker;
end;

constructor TWorker.Create(ATCPBase: TTCPBase; AStackSize: SizeUInt;
  ASocket: TSocket);
begin
  inherited Create(ATCPBase, AStackSize);
  if FMaxThreadException then
    Exit;
  Randomize;
  FForceQuit := False;
  FCnt := 0;
  FTotCnt := 0;
  FSpeed := 0;
  FRemaining := 0;
  FFileName := '';
  FMS := TMemoryStream.Create;
  FTCPBase := ATCPBase;
  FConnection := TConnection.Create;
  FThreadType := ttAcceptedWorker;
  FBlockSocket.Socket := ASocket;
end;

destructor TWorker.Destroy;
begin
  if FMaxThreadException then
    Exit;
  FMS.Free;
  FConnection.Free;
  inherited Destroy;
end;

procedure TWorker.Upload;
var
  Res: Boolean;
  Msg: String;
  Params: TStringArray;
  Size: Int64;
  FileName: String;
  Path: String;
begin
  Res := False;
  FileName := '';
  case FThreadType of
    ttConnectedWorker:
      begin
        FMsg := 'FILE';
        SetLength(FParams, Length(FParams) + 2);
        FParams[Length(FParams) - 2] := IntToStr(FileSize(FFileName));
        FParams[Length(FParams) - 1] := ExtractFileName(FFileName);
        if SendMessage(FMsg, FParams, True) then
          if SendFile(FFileName) then
          begin
            Res := True;
            UpdateGUI(@DoOnSentFile);
          end;
      end;
    ttAcceptedWorker:
      begin
        if not RecvMessage(5000, Msg, Params) then
          Exit;
        Size := StrToInt64Def(Params[Length(Params) - 2], -1);
        FileName := Params[Length(Params) - 1];
        SetLength(Params, Length(Params) - 2);
        if Size > 0 then
        begin
          Path := AddDirSeparator(GetDownloadDir(FTCPBase.FDownloadDirectory));
          if DirectoryExists(Path) then
          begin
            FileName := ChangeFileName(FileName, FConData.FRemoteUser, FConData.FLocalUser);
            Path := Path + FileName;
            if RecvFile(Path, Size) then
            begin
              Res := True;
              UpdateGUI(@DoOnRecvFile);
            end;
          end
          else
            ErrorMessage(rsInvalidDirectory, 0);
        end
        else
          FileCreate(Params[0] + Params[1]);
      end;
  end;
  if (not Res) or (FForceQuit) or (FBlockSocket.LastError <> 0) then
    Progress(FileName, -1, -1, True);
end;

procedure TWorker.Download;
var
  Res: Boolean;
  Msg: String;
  Params: TStringArray;
  Size: Int64;
  FileName: String;
  Path: String;
begin
  Res := False;
  FileName := '';
  case FThreadType of
    ttConnectedWorker:
      begin
        if not SendMessage('FILE', [FFileName], True) then
          Exit;
        if not RecvMessage(5000, Msg, Params) then
          Exit;
        Size := StrToInt64Def(Params[Length(Params) - 2], -1);
        FileName := Params[Length(Params) - 1];
        SetLength(Params, Length(Params) - 2);
        if Size > 0 then
        begin
          Path := AddDirSeparator(GetDownloadDir(FTCPBase.FDownloadDirectory));
          if DirectoryExists(Path) then
          begin
            FileName := ChangeFileName(FileName, FConData.FRemoteUser, FConData.FLocalUser);
            Path := Path + FileName;
            if RecvFile(Path, Size) then
            begin
              Res := True;
              UpdateGUI(@DoOnRecvFile);
            end;
          end
          else
            ErrorMessage(rsInvalidDirectory, 0);
        end
        else
          FileCreate(Params[0] + Params[1]);
      end;
    ttAcceptedWorker:
      begin
        if not RecvMessage(5000, Msg, Params) then
          Exit;
        FWorkDirection := wdUpload;
        FileName := Params[0];
        if FileExists(FileName) then
        begin
          SetLength(FParams, Length(FParams) + 2);
          FParams[Length(FParams) - 2] := IntToStr(FileSize(FileName));
          FParams[Length(FParams) - 1] := ExtractFileName(FileName);
          if SendMessage('FILE', FParams, True) then
          begin
            if SendFile(FileName) then
            begin
              Res := True;
              UpdateGUI(@DoOnSentFile);
            end;
          end;
        end
        else
          ErrorMessage(Format(rsFileNotExists, [Msg]), 0);
      end;
  end;
  if (not Res) or (FForceQuit) or (FBlockSocket.LastError <> 0) then
    Progress(FileName, -1, -1, True);
end;

procedure TWorker.Execute;
var
  Msg: String;
  Params: TStringArray;
  WorkDirection: TWorkDirection;
begin
  FreeOnTerminate := True;
  FStarted := True;
  FTick := GetTickCount64;
  try
    UpdateThreadList(utAdd);
    case FThreadType of
      ttConnectedWorker:
        begin
          if not RecvMessage(5000, Msg, Params) or (Trim(Msg) <> 'READY') then
            Exit;
          if not SendMessage(Ord(FWorkDirection).ToString, [], True) then
            Exit;
          WorkDirection := FWorkDirection;
        end;
      ttAcceptedWorker:
        begin
          if not SendMessage('READY', [], True) then
            Exit;
          if not RecvMessage(5000, Msg, FParams) then
            Exit;
          WorkDirection := TWorkDirection(StrToIntDef(Msg, 0));
        end;
    end;
    case WorkDirection of
      wdUpload: Upload;
      wdDownload: Download;
    end;
  finally
    UpdateThreadList(utRemove);
  end;
end;

procedure TWorker.DoOnSentFile;
begin
  if Assigned(FTCPBase.FOnSentFile) then
    FTCPBase.FOnSentFile(Self, FFileName, FConData);
end;

procedure TWorker.DoOnRecvFile;
begin
  if Assigned(FTCPBase.FOnRecvFile) then
    FTCPBase.FOnRecvFile(Self, FFileName, FConData);
end;

procedure TWorker.DoOnFileProgress;
begin
  if Assigned(FTCPBase.FOnFileProgress) then
    FTCPBase.FOnFileProgress(Self, FWorkDirection, FFileName, FCnt, FTotCnt, FSpeed, FRemaining, FConData);
end;

procedure TWorker.Progress(AFileName: String; ACnt, ATotCnt: Int64;
  AError: Boolean);
var
  ElapsedSec: Single;
  Speed: LongInt;
  Remaining: LongInt;
  Tick: QWord;
  Diff: QWord;
begin
  if (not FTCPBase.FActiveOrConnected) or (Terminated) then
    Exit;

  if AError then
  begin
    FCnt := ACnt;
    FTotCnt := ATotCnt;
    UpdateGUI(@DoOnFileProgress);
    Exit;
  end;

  Tick := GetTickCount64;
  Diff  := Tick - FTick;
  if (Diff > 1000) or (ACnt >= ATotCnt) then
  begin
    ElapsedSec := Diff/1000;
    Speed := Trunc((ACnt - FOldCnt)/ElapsedSec);
    if Speed > 0 then
      Remaining := Round((ATotCnt - ACnt)/Speed);
    FTick := Tick;
    FOldCnt := ACnt;
    FSpeed := Speed;
    FFileName := AFileName;
    FRemaining := Remaining;
    FCnt := ACnt;
    FTotCnt := ATotCnt;
    UpdateGUI(@DoOnFileProgress, True);
  end;
end;

function TWorker.SendStream(AMS: TMemoryStream): Boolean;
var
  ReadCnt: LongInt;
  Cnt: Int64;
  Buffer: Pointer;
  Ms: TMemoryStream;
  MsComp: TMemoryStream;
  Msg: String;
  CompRate: Single;
begin
  Result := False;
  if not FTCPBase.FActiveOrConnected then
    Exit;
  if FBlockSocket.LastError > 0 then
  begin
    ErrorMessage(Format(rsStreamSentError, [FConData.FRemoteIP]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
    Exit;
  end;
  FOldCnt := 0;
  MS := TMemoryStream.Create;
  try
    AMS.Position := 0;
    MSComp := TMemoryStream.Create;
    try
      GetMem(Buffer, FBlockSocket.SendMaxChunk);
      try
        Cnt := 0;
        repeat
          if (not FTCPBase.FActiveOrConnected) or (FForceQuit) then
            Exit;
          Msg :=  FBlockSocket.RecvString(5000);
          if (FBlockSocket.LastError > 0) or (Msg = 'DONE') or (Msg <> 'NEXT') then
            Break;
          ReadCnt := AMS.Read(Buffer^, FBlockSocket.SendMaxChunk);
          if (ReadCnt > 0) then
          begin
            MS.Clear;
            MS.Write(Buffer^, ReadCnt);
            MS.Position := 0;
            if FTCPBase.FCompress and  CompressStream(Ms, MsComp, clBestCompression, CompRate) then
              FBlockSocket.SendStream(MSComp)
            else
              FBlockSocket.SendStream(MS);
            if (FBlockSocket.LastError > 0) and (FBlockSocket.LastError <> WSAETIMEDOUT) then
              Break;
            if (FBlockSocket.LastError = 0) and (FTCPBase.FActiveOrConnected) then
            begin
              Cnt := Cnt + MS.Size;
              Progress('', Cnt, AMS.Size);
            end;
          end;
        until (ReadCnt = 0) or (Cnt > AMS.Size);
        Result := (FBlockSocket.LastError = 0) and (Cnt >= AMS.Size);
        if not Result then
          ErrorMessage(Format(rsStreamSentError, [FConData.FRemoteIP]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
      finally
        FreeMem(Buffer);
      end;
    finally
      MsComp.Free;
    end;
  finally
    MS.Free;
  end;
end;

function TWorker.RecvStream(AMS: TMemoryStream): Boolean;
var
  Cnt: Int64;
  MS: TMemoryStream;
  MSDecomp: TMemoryStream;
begin
  Result := False;
  if not FTCPBase.FActiveOrConnected then
    Exit;
  MS := TMemoryStream.Create;
  try
    Cnt := 0;
    MSDecomp := TMemoryStream.Create;
    try
      repeat
        if (not FTCPBase.FActiveOrConnected) or (FForceQuit) then
          Exit;
        FBlockSocket.SendString('NEXT' + #13#10);
        MS.Clear;
        FBlockSocket.RecvStream(Ms, 5000);
        if (FBlockSocket.LastError > 0) and (FBlockSocket.LastError <> WSAETIMEDOUT) then
          Break;
        if (MS.Size > 0) and (FBlockSocket.LastError = 0) and (FTCPBase.FActiveOrConnected) then
        begin
          MS.Position := 0;
          if FTCPBase.FCompress and DecompressStream(MS, MSDecomp) then
          begin
            AMS.CopyFrom(MSDecomp, MSDecomp.Size);
            Cnt := Cnt + MSDecomp.Size;
            Progress('', Cnt, AMS.Size);
          end
          else
          begin
            AMS.CopyFrom(MS, MS.Size);
            Cnt := Cnt + MS.Size;
            Progress('', Cnt, AMS.Size);
          end;
        end;
      until (Cnt >= AMS.Size);
      FBlockSocket.SendString('DONE' + #13#10);
      Result := (FBlockSocket.LastError = 0) and (Cnt >= AMS.Size);
      if not Result then
        ErrorMessage(Format(rsStreamRecvError, [FConData.FRemoteIP]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
    finally
      MSDecomp.Free;
    end;
  finally
    MS.Free;
  end
end;

function TWorker.SendFile(AFileName: String): Boolean;
var
  ReadCnt: LongInt;
  Cnt: Int64;
  Buffer: Pointer;
  MS: TMemoryStream;
  MSComp: TMemoryStream;
  FS: TFileStream;
  Msg: String;
  CompRate: Single;
begin
  Result := False;
  if not FTCPBase.FActiveOrConnected then
    Exit;
  if FBlockSocket.LastError > 0 then
  begin
    ErrorMessage(Format(rsFileSentError, [FConData.FRemoteIP]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
    Exit;
  end;
  try
    FS := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyNone);
    try
      FS.Position := 0;
      GetMem(Buffer, FBlockSocket.SendMaxChunk);
      try
        Cnt := 0;
        MS := TMemoryStream.Create;
        try
          MSComp := TMemoryStream.Create;
          try
            repeat
              if (not FTCPBase.FActiveOrConnected) or (FForceQuit) then
                Exit;
              Msg := FBlockSocket.RecvString(5000);
              if (FBlockSocket.LastError > 0) or (Msg = 'DONE') or (Msg <> 'NEXT') then
                Break;
              ReadCnt := FS.Read(Buffer^, FBlockSocket.SendMaxChunk);
              if (ReadCnt > 0) then
              begin
                MS.Clear;
                MS.Write(Buffer^, ReadCnt);
                MS.Position := 0;
                if FTCPBase.FCompress and CompressStream(Ms, MsComp, clBestCompression, CompRate) then
                  FBlockSocket.SendStream(MSComp)
                else
                  FBlockSocket.SendStream(MS);
                if (FBlockSocket.LastError > 0) and (FBlockSocket.LastError <> WSAETIMEDOUT) then
                  Break;
                if (FBlockSocket.LastError = 0) and (FTCPBase.FActiveOrConnected) then
                begin
                  Cnt := Cnt + MS.Size;
                  Progress(AFileName, Cnt, FS.Size);
                end;
              end;
            until (ReadCnt = 0) or (Cnt > FS.Size);
            Result := (FBlockSocket.LastError = 0) and (Cnt >= FS.Size);
            if not Result then
              ErrorMessage(Format(rsFileSentError, [FConData.FRemoteIP]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
          finally
            MSComp.Free;
          end;
        finally
          MS.Free;
        end;
      finally
        FreeMem(Buffer);
      end;
    finally
      FS.Free;
    end;
  except
    on E: Exception do
      ErrorMessage(Format(rsFileSentError, [FConData.FRemoteIP]) + E.Message, 0);
  end
end;

function TWorker.RecvFile(AFileName: String; ASize: Int64): Boolean;
var
  Cnt: Int64;
  MS: TMemoryStream;
  MSDecomp: TMemoryStream;
  FS: TFileStream;
begin
  Result := False;
  if not FTCPBase.FActiveOrConnected then
    Exit;
  try
    FS := TFileStream.Create(AFileName, fmCreate or fmOpenReadWrite or fmShareDenyWrite);
    try
      MS := TMemoryStream.Create;
      try
        Cnt := 0;
        MSDecomp := TMemoryStream.Create;
        try
          repeat
            if (not FTCPBase.FActiveOrConnected) or (FForceQuit) then
              Exit;
            FBlockSocket.SendString('NEXT' + #13#10);
            MS.Clear;
            FBlockSocket.RecvStream(MS, 5000);
            if (FBlockSocket.LastError > 0) and (FBlockSocket.LastError <> WSAETIMEDOUT) then
              Break;
            if (MS.Size > 0) and (FBlockSocket.LastError = 0) and (FTCPBase.FActiveOrConnected) then
            begin
              MS.Position := 0;
              if FTCPBase.FCompress and DecompressStream(MS, MSDecomp) then
              begin
                 FS.CopyFrom(MsDecomp, MSDecomp.Size);
                 Cnt := Cnt + MSDecomp.Size;
                 Progress(AFileName, Cnt, ASize);
              end
              else
              begin
                FS.CopyFrom(MS, MS.Size);
                Cnt := Cnt + MS.Size;
                Progress(AFileName, Cnt, ASize);
              end;
            end;
          until (Cnt >= ASize);
          FBlockSocket.SendString('DONE' + #13#10);
          Result := (FBlockSocket.LastError = 0) and (Cnt >= ASize);
          if not Result then
            ErrorMessage(Format(rsFileRecvError, [FConData.FRemoteIP]) + FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
        finally
          MSDecomp.Free;
        end;
      finally
        MS.Free;
      end
    finally
      FS.Free;
    end;
  except
    on E: Exception do
      ErrorMessage(Format(rsFileRecvError, [FConData.FRemoteIP]) + E.Message, 0);
  end
end;

function TWorker.AddDirSeparator(const Path: String): String;
begin
  if (Trim(Path) <> '') and (not (Path[Length(Path)] in AllowDirectorySeparators)) then
    Result := Path + PathDelim
  else
    Result := Path;
end;

function TWorker.GetDownloadDir(const APath: String): String;
begin
  if Trim(APath) = '' then
  begin
    Result := AddDirSeparator(GetUserDir) + 'Downloads';
    Exit;
  end;
  Result := AddDirSeparator(APath);
  if not DirectoryExists(Result) then
  begin
    if not CreateDir(Result) then
      Result := AddDirSeparator(GetUserDir) + 'Downloads';
  end
end;

function TWorker.ChangeFileName(AFileName, AFrom, ATo: String): String;
var
  Ext: String;
  Name: String;
begin
  Ext := ExtractFileExt(AFileName);
  Name := ChangeFileExt(AFileName, '');
  Result := Name + '[' + AFrom + '_' + ATo  + ']' + Ext;
end;

{ TTCPBase }
constructor TTCPBase.Create;
begin
  FUniqueID := 0;
  FCompress := False;
  FIsDaemon := False;
  FActiveOrConnected := False;
  FShowErrorMessages := False;
  FShowInternalMessages := False;
  FLastErrCode := 0;
  FLastErrMsg := '';
  FThreadList := TLockList.Create;
end;

destructor TTCPBase.Destroy;
begin
  FThreadList.Free;
  inherited Destroy;
end;

procedure TTCPBase.CopyParams(ASrc: TStringArray; out ADst: TStringArray);
begin
  ADst := nil;
  if Length(ASrc) > 0 then
  begin
    ADst := ASrc;
    SetLength(ADst, Length(ASrc));
  end;
end;

procedure TTCPBase.CopyStream(ASrc, ADst: TMemoryStream);
begin
  if ASrc.Size > 0 then
  begin
    ASrc.Position := 0;
    ADst.CopyFrom(ASrc, ASrc.Size);
    ADst.Position := 0;
  end;
end;

procedure TTCPBase.ErrorMessage(AThread: TBase; AErrMsg: String;
  AErrCode: Integer; AConData: TConData);
begin
  FLastErrMsg := AErrMsg;
  FLastErrCode := AErrCode;
end;

{ TClientCon }
constructor TConnection.Create;
begin
  FIP := '';
  FPort := DefPort;
  FUser := '';
  FPass := '';
  FData := nil;
  FProxy.FProxyType := ptNone;
  FProxy.FProxyIP := '';
  FProxy.FProxyPort := DefProxyPort;
  FProxy.FProxyUser := '';
  FProxy.FProxyPass := '';
end;

{ TConnectedClient }
procedure TConnectedClient.DoOnConnected;
begin
  if Assigned(FTCPClient.FOnConnected) then
    FTCPClient.FOnConnected(Self, FConData);
end;

procedure TConnectedClient.DoOnDisconnected;
begin
  if Assigned(FTCPClient.FOnDisconnected) then
    FTCPClient.FOnDisconnected(Self, FConData);
end;

procedure TConnectedClient.DoOnRecvMessage;
begin
  if Assigned(FTCPClient.FOnRecvMessage) then
    FTCPClient.FOnRecvMessage(Self, FMsg, FParams, FConData);
end;

procedure TConnectedClient.DoOnFileRequest;
begin
  if Trim(FParams[0]) <> '' then
    FTCPClient.FileRequest(FParams[0]);
end;

procedure TConnectedClient.Execute;
var
  Msg: String;
  Params: TStringArray;
  CurPing: QWord;
begin
  FStarted := True;
  FreeOnTerminate := True;
  FLastPing := GetTickCount64;
  UpdateThreadList(utAdd);
  UpdateGUI(@DoOnConnected);
  try
    while FTCPClient.FActiveOrConnected do
    begin
      if (FForceQuit) or (not RecvMessage(500, Msg, Params)) then
        Break;
      if Trim(Msg) <> '' then
        case Msg of
          'FILE':
            begin
              FTCPBase.CopyParams(Params, FParams);
              UpdateGUI(@DoOnFileRequest);
            end;
          else
            begin
              FMsg := Msg;
              FTCPBase.CopyParams(Params, FParams);
              UpdateGUI(@DoOnRecvMessage);
              FLastPing := GetTickCount64;
            end;
        end;

      CurPing := GetTickCount64;
      if (CurPing - FLastPing > PingTimeoutClient) then
      begin
        FLastPing := CurPing;
        SendMessage('PING', [], True);
      end;
    end;
    while FBusy do
      Wait(100);
  finally
    FBlockSocket.CloseSocket;
    UpdateGUI(@DoOnDisconnected, True);
    UpdateThreadList(utRemove);
    while GetThreadCnt > 0 do
      Wait(100);
    OnTerminate(Self);
  end;
end;

constructor TConnectedClient.Create(ATCPBase: TTCPBase; AStackSize: SizeUInt);
begin
  inherited Create(ATCPBase, AStackSize);
  if FMaxThreadException then
    Exit;
  FForceQuit := False;
  FThreadType := ttConnectedClient;
  FTCPClient := TTCPClient(ATCPBase);
end;

destructor TConnectedClient.Destroy;
begin
  if FMaxThreadException then
    Exit;
  inherited Destroy;
end;

{ TTCPClient }
constructor TTCPClient.Create;
begin
  inherited Create;
  FConnection := TConnection.Create;
end;

destructor TTCPClient.Destroy;
begin
  FConnection.Free;
  inherited Destroy;
end;

function TTCPClient.GetConnected: Boolean;
begin
  Result := FActiveOrConnected and (not FBusy);
end;

procedure TTCPClient.ErrorMessage(AThread: TBase; AErrMsg: String;
  AErrCode: Integer; AConData: TConData);
begin
  inherited ErrorMessage(AThread, AErrMsg, AErrCode, AConData);
  if (not FActiveOrConnected) or (not FShowErrorMessages) then
    Exit;
  if Assigned(FOnSocketError) then
    FOnSocketError(AThread, AErrMsg, AErrCode, AConData);
end;

function TTCPClient.Connect(ABase: TBase; AConnection: TConnection): Boolean;
begin
  Result := False;
  ABase.FBlockSocket.Connect(AConnection.FIP, AConnection.FPort.ToString);
  if FUseProxy then
  begin
    if (AConnection.FProxy.FProxyType = ptSocks5) or (AConnection.FProxy.FProxyType = ptSocks4) then
    begin
      if ABase.FBlockSocket.LastError = 0 then
      begin
        ABase.FBlockSocket.RecvPacket(TunnelTimeout);
        Result := (ABase.FBlockSocket.LastError = 0) or ((ABase.FBlockSocket.LastError > 0)  and (ABase.FBlockSocket.LastError = WSAETIMEDOUT));
      end;
    end
    else
      Result := (ABase.FBlockSocket.LastError = 0);
  end
  else
    Result := (ABase.FBlockSocket.LastError = 0);;
  if Result then
    Result := ABase.GetExtendedInfo
  else
    ErrorMessage(ABase, ABase.FBlockSocket.LastErrorDesc, ABase.FBlockSocket.LastError, ABase.FConData);
end;

function TTCPClient.Login(ABase: TBase; AConnection: TConnection): Boolean;
var
  Msg: String;
  Params: TStringArray;
begin
  Result := False;
  if not ABase.SendMessage('LOGIN', [AConnection.FPass, Ord(ABase.FThreadType).ToString, AConnection.FUser], True) then
    Exit;
  if ABase.RecvMessage(Timeout, Msg, Params) then
  begin
    if (Msg = 'LOGINREPLY') then
    begin
      case Params[0] of
        'SUCCESS':
          begin
            Result := True;
            ABase.FConData.FData := FConnection.FData;
            ABase.FConData.FLocalUser := FConnection.FUser;
            ABase.FConData.FRemoteUser := Params[1];
          end;
        'FAILEDINVALID':
          ErrorMessage(ABase, rsInvalidUserNameOrPassword, 10013, ABase.FConData);
        'FAILEDDUPLICATE':
          ErrorMessage(ABase, rsInvalidDuplicateUser, 10013, ABase.FConData);
      end;
    end;
  end;
end;

procedure TTCPClient.SetupProxy;
begin
  with FCC, FConnection.FProxy do
  begin
    case FProxyType of
      ptHTML:
        begin
          FBlockSocket.HTTPTunnelIP := FProxyIP;
          FBlockSocket.HTTPTunnelPort := FProxyPort.ToString;
          FBlockSocket.HTTPTunnelUser := FProxyUser;
          FBlockSocket.HTTPTunnelPass := FProxyPass;
          FBlockSocket.HTTPTunnelTimeout := TunnelTimeout;
        end;
      ptSocks4, ptSocks5:
        begin
          if FProxyType = ptSocks4 then
            FBlockSocket.SocksType := ST_Socks4
          else
            FBlockSocket.SocksType := ST_Socks5;
          FBlockSocket.SocksIP:= FProxyIP;
          FBlockSocket.SocksPort := FProxyPort.ToString;
          FBlockSocket.SocksUsername := FProxyUser;
          FBlockSocket.SocksPassword := FProxyPass;
          FBlockSocket.SocksTimeout := TunnelTimeout;
        end;
      ptNone:
        begin
          FBlockSocket.HTTPTunnelIP := '';
          FBlockSocket.SocksIP := '';
        end;
    end;
  end;
end;

procedure TTCPClient.CopyConnection(ASrc, ADst: TConnection);
begin
  with ADst do
  begin
    FIP := ASrc.FIP;
    FPort := ASrc.FPort;
    FUser := ASrc.FUser;
    FPass := ASrc.FPass;
    FData := ASrc.FData;
    FProxy.FProxyType := ASrc.FProxy.FProxyType;
    FProxy.FProxyIP := ASrc.FProxy.FProxyIP;
    FProxy.FProxyPort := ASrc.FProxy.FProxyPort;
    FProxy.FProxyUser := ASrc.FProxy.FProxyUser;
    FProxy.FProxyPass := ASrc.FProxy.FProxyPass;
  end;
end;

procedure TTCPClient.DoOnTerminate(Sender: TObject);
begin
  FCC := nil;
  FActiveOrConnected := False;
end;

procedure TTCPClient.FileRequest(AFileName: String);
begin
  if (not FActiveOrConnected) or (FCC = nil) then
    Exit;
  StartWorker(wdDownload, AFileName, nil);
end;

procedure TTCPClient.Connect(const AConnection: TConnection);
begin
  if (FActiveOrConnected) or (FBusy) then
    Exit;
  FBusy := True;
  try
    CopyConnection(AConnection, FConnection);
    FUseProxy := FConnection.FProxy.FProxyType <> ptNone;
    if FUseProxy then
      SetupProxy;
    FCC := TConnectedClient.Create(Self, BaseThreadStackSize);
    FCC.OnTerminate := @DoOnTerminate;
    if Connect(FCC, FConnection) then
    begin
      FActiveOrConnected := True;
      if Login(FCC, FConnection) then
      begin
        FCC.Start;
        Exit;
      end
      else
        FActiveOrConnected := False;
    end;
    FCC.FBlockSocket.CloseSocket;
    FCC.Free;
    FCC := nil;
  finally
    FBusy := False;
  end;
end;

procedure TTCPClient.Disconnect;
begin
  if (not FActiveOrConnected) or (FBusy) then
    Exit;
  FBusy := True;
  try
    FActiveOrConnected := False;
    while FCC <> nil do
      Sleep(100);
  finally
    FBusy := False;
  end;
end;

function TTCPClient.SendMessage(AMsg: String; AParams: TStringArray;
  AData: Pointer): Boolean;
begin
  Result := False;
  if (not FActiveOrConnected) or (FCC = nil) then
    Exit;
  Result := FCC.SendMessage(AMsg, AParams);
end;

function TTCPClient.StartWorker(AWorkDirection: TWorkDirection; AFileName: String;
  AData: Pointer): Boolean;
var
  Worker: TWorker;
begin
  Result := False;
  Worker := TWorker.Create(Self, WorkerThreadStackSize);
  Worker.FWorkDirection := AWorkDirection;
  Worker.FThreadType := ttConnectedWorker;
  CopyConnection(FConnection, Worker.FConnection);
  if Connect(Worker, Worker.FConnection) then
  begin
    if Login(Worker, Worker.FConnection) then
    begin
      Worker.FFileName := AFileName;
      Worker.FConData.FData := AData;
      Worker.Start;
      Result := True;
      Exit;
    end;
  end;
  Worker.Free;
end;

function TTCPClient.SendFile(AFileName: String; AData: Pointer): Boolean;
begin
  Result := False;
  if (not FActiveOrConnected) or (FCC = nil) then
    Exit;
  Result := StartWorker(wdUpload, AFileName, AData);
end;

{ TBindCon }
constructor TBind.Create;
begin
  FIP := '';
  FPort := DefPort;
  FUser := '';
  FPass := '';
  FData := nil;
end;

{ TAcceptedClient }
procedure TAcceptedClient.DoOnConnected;
begin
  if Assigned(FTCPServer.FOnConnected) then
    FTCPServer.FOnConnected(Self, FConData, FConCnt);
end;

procedure TAcceptedClient.DoOnDisconnected;
begin
  if Assigned(FTCPServer.FOnDisconnected) then
    FTCPServer.FOnDisconnected(Self, FConData, FConCnt, FForceQuit);
end;

procedure TAcceptedClient.DoOnRecvMessage;
begin
  if Assigned(FTCPServer.FOnRecvMessage) then
    FTCPServer.FOnRecvMessage(Self, FMsg, FParams, FConData);
end;

function TAcceptedClient.AddUser(AUserName: String): Boolean;
var
  UserHash: ShortString;
  UserData: PUserData;
begin
  Result := True;
  while (not Terminated) do
  begin
    if FTCPServer.FLock.Lock then
    begin
      try
        UserHash := MD5Print(MD5String(LowerCase(AUserName)));
        Result := FTCPServer.FUserList.FindIndexOf(UserHash) = -1;
        if Result then
        begin
          New(UserData);
          UserData^.FName := AUserName;
          UserData^.FIP := FConData.FRemoteIP;
          FTCPServer.FUserList.Add(UserHash, UserData);
          FConCnt := FTCPServer.FUserList.Count;
        end;
      finally
        FTCPServer.FLock.Unlock;
      end;
      Break;
    end;
    Wait(Random(100));
  end;
end;

function TAcceptedClient.RemoveUser(AUserName: String): Boolean;
var
  UserHash: ShortString;
  UserData: PUserData;
begin
  Result := False;
  while (not Terminated) do
  begin
    if FTCPServer.FLock.Lock then
    begin
      try
        UserHash := MD5Print(MD5String(LowerCase(AUserName)));
        UserData := FTCPServer.FUserList.Find(UserHash);
        if UserData <> nil then
        begin
          FTCPServer.FUserList.Remove(UserData);
          Dispose(UserData);
          FConCnt := FTCPServer.FUserList.Count;
        end;
      finally
        FTCPServer.FLock.Unlock;
      end;
      Break;
    end;
    Wait(Random(100));
  end;
end;

procedure TAcceptedClient.StartWorker;
var
  Worker: TWorker;
  Socket: TSocket;
begin
  Socket := FBlockSocket.Socket;
  FBlockSocket.Socket := 0;
  Worker := TWorker.Create(FTCPBase, WorkerThreadStackSize, Socket);
  FTCPServer.CopyConnectionData(FConData, Worker.FConData);
  Worker.Start;
end;

function TAcceptedClient.Login: Byte;
var
  Msg: String;
  Params: TStringArray;
  ThreadType: TThreadType;
begin
  Result := 0;
  if RecvMessage(Timeout, Msg, Params) then
  begin
    if (Msg = 'LOGIN') and (Length(Params) = 3) and (Params[0] = Trim(FTCPServer.FBind.FPass)) then
    begin
      ThreadType := TThreadType(StrToIntDef(Params[1], 0));
      case ThreadType of
        ttConnectedClient:
          begin
            if AddUser(Params[2]) then
            begin
              Result := 1;
              FConData.FLocalUser := FTCPServer.FBind.FUser;
              FConData.FRemoteUser := Params[2];
              SendMessage('LOGINREPLY', ['SUCCESS', FConData.FLocalUser], True);
            end
            else
              SendMessage('LOGINREPLY', ['FAILEDDUPLICATE'], True);
          end;
        ttConnectedWorker:
          begin
            Result := 2;
            FConData.FLocalUser := FTCPServer.FBind.FUser;
            FConData.FRemoteUser := Params[2];
            SendMessage('LOGINREPLY', ['SUCCESS', FConData.FLocalUser], True);
          end;
      end;
    end
    else
      SendMessage('LOGINREPLY', ['FAILEDINVALID'], True);
  end;
end;

procedure TAcceptedClient.Execute;
var
  Msg: String;
  Params: TStringArray;
  CurPing: QWord;
begin
  FreeOnTerminate := True;
  case Login of
    0: Exit;
    1: FStarted := True;
    2: begin
         StartWorker;
         Exit;
       end;
  end;
  CurPing := GetTickCount64;
  try
    UpdateThreadList(utAdd);
    UpdateGUI(@DoOnConnected);
    while FTCPServer.FActiveOrConnected do
    begin
      if (FForceQuit) or (not RecvMessage(500, Msg, Params)) then
       Break;

      if Trim(Msg) <> '' then
      begin
        case Msg of
          'PING':
            begin
              FLastPing := GetTickCount64;
            end
          else
            begin
              FMsg := Msg;
              FTCPBase.CopyParams(Params, FParams);
              UpdateGUI(@DoOnRecvMessage);
              FLastPing := GetTickCount64;
            end;
        end;
      end;

      CurPing := GetTickCount64;
      if (CurPing - FLastPing > PingTimeoutServer) then
        Break;
    end;
    while FBusy do
      Wait(100);
  finally
    FBlockSocket.CloseSocket;
    if FTCPServer.FActiveOrConnected then
    begin
      RemoveUser(FConData.FRemoteUser);
      UpdateGUI(@DoOnDisconnected);
    end;
    UpdateThreadList(utRemove);
  end;
end;

constructor TAcceptedClient.Create(ATCPBase: TTCPBase; AStackSize: SizeUInt;
  ASocket: TSocket);
begin
  inherited Create(ATCPBase, AStackSize);
  if FMaxThreadException then
    Exit;
  FBlockSocket.Socket := ASocket;
  FForceQuit := False;
  FThreadType := ttAcceptedClient;
  FLastPing := GetTickCount64;
  FTCPServer := TTCPServer(ATCPBase);
end;

destructor TAcceptedClient.Destroy;
begin
  if FMaxThreadException then
    Exit;
  inherited Destroy;
end;

{ TThreadManager }
constructor TThreadManager.Create(ATCPBase: TTCPBase);
begin
  inherited Create(ATCPBase, BaseThreadStackSize);
  FTCPServer := TTCPServer(ATCPBase);
end;

procedure TThreadManager.Execute;
var
  AC: TAcceptedClient;
  Socket: TSocket;
  ThreadCnt: integer;
begin
  FStarted := True;
  while FTCPServer.FActiveOrConnected do
  begin
    if FBlockSocket.CanRead(1000) then
    begin
      Socket := FBlockSocket.Accept;
      if FBlockSocket.LastError = 0 then
      begin
        AC := TAcceptedClient.Create(FTCPServer, BaseThreadStackSize, Socket);
        if not AC.FMaxThreadException then
        begin
          if AC.GetExtendedInfo then
            AC.Start
          else
            ErrorMessage(FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
        end
      end
      else
        ErrorMessage(FBlockSocket.LastErrorDesc, FBlockSocket.LastError);
    end;
  end;
  ThreadCnt := -1;
  repeat
    ThreadCnt := GetThreadCnt;
    Sleep(100);
  until ThreadCnt = 0;
end;

{ TTCPServer }
constructor TTCPServer.Create;
begin
  inherited Create;
  FBusy := False;
  FLock := TLock.Create;
  FBind := TBind.Create;
  FUserList := TFPHashList.Create;
end;

destructor TTCPServer.Destroy;
begin
  ClearUserList;
  FUserList.Free;
  FBind.Free;
  FLock.Free;
  inherited Destroy;
end;

function TTCPServer.GetActive: Boolean;
begin
  Result := FActiveOrConnected and (not FBusy);
end;

procedure TTCPServer.ClearUserList;
var
  I: Integer;
  UserData: PUserData;
begin
  for I := FUserList.Count - 1 downto 0 do
  begin
    UserData := FUserList.Items[I];
    Dispose(UserData);
  end;
  FUserList.Clear;
end;

procedure TTCPServer.ErrorMessage(AThread: TBase; AErrMsg: String;
  AErrCode: Integer; AConData: TConData);
begin
  inherited ErrorMessage(AThread, AErrMsg, AErrCode, AConData);
  if (not FActiveOrConnected) or (not FShowErrorMessages) then
    Exit;
  if Assigned(FOnSocketError) then
    FOnSocketError(AThread, AErrMsg, AErrCode, AConData);
end;

procedure TTCPServer.CopyBind(ASrc, ADst: TBind);
begin
  with ADst do
  begin
    FIP := ASrc.FIP;
    FPort := ASrc.FPort;
    FUser := ASrc.FUser;
    FPass := ASrc.FPass;
  end;
end;

procedure TTCPServer.CopyConnectionData(ASrc, ADst: TConData);
begin
  with ADst do
  begin
    FUniqueID := ASrc.FUniqueID;
    FDateTime := ASrc.FDateTime;
    FLocalIP := ASrc.FLocalIP;
    FLocalPort := ASrc.FLocalPort;
    FLocalHost := ASrc.FLocalHost;
    FLocalUser := ASrc.FLocalUser;
    FRemoteIP := ASrc.FRemoteIP;
    FRemotePort := ASrc.FRemotePort;
    FRemoteHost := ASrc.FRemoteHost;
    FRemoteUser := ASrc.FRemoteUser;
    FData := ASrc.FData;
  end;
end;

procedure TTCPServer.Start(ABindCon: TBind);
begin
  if (FActiveOrConnected) or (FBusy) then
    Exit;
  FBusy := True;
  try
    CopyBind(ABindCon, FBind);
    FTM := TThreadManager.Create(Self);
    with FTM.FBlockSocket do
    begin
      CreateSocket;
      SetLinger(True, 10000);
      Bind(FBind.FIP, FBind.FPort.ToString);
      Listen;
      if LastError = 0 then
      begin
        FActiveOrConnected := True;
        FTM.Start;
      end
      else
      begin
        ErrorMessage(nil, FTM.FBlockSocket.LastErrorDesc, FTM.FBlockSocket.LastError, FTM.FConData);
        FTM.Free;
        FTM := nil;
      end;
    end;
  finally
    FBusy := False;
  end;
end;

procedure TTCPServer.Stop;
begin
  if (not FActiveOrConnected) or (FBusy) then
    Exit;
  FBusy := True;
  try
    FActiveOrConnected := False;
    FTM.Terminate;
    FTM.WaitFor;
    FTM.Free;
    FTM := nil;
    ClearUserList;
    FThreadList.FList.Clear;
  finally
    FBusy := False;
  end;
end;

function TTCPServer.SendMessage(AAC: TAcceptedClient; AMsg: String;
  AParams: TStringArray): Boolean;
begin
  Result := False;
  if (not FActiveOrConnected) then
    Exit;
  Result := AAC.SendMessage(AMsg, AParams);
end;

function TTCPServer.SendFile(AAC: TAcceptedClient; AFileName: String): Boolean;
begin
  Result := False;
  if (not FActiveOrConnected) then
    Exit;
  Result := AAC.SendMessage('FILE', [AFileName], True);
end;

end.

